# immortal_letters + a total rewrite of growing_letters (except for docs).
# https://trac.sagemath.org/ticket/31760

def immortal_letters(self):
    r"""
    Return the set of immortal letters.

    A letter `a` is *immortal* for the morphism `s` if the length of the
    iterates of `| s^n(a) |` is larger than zero as `n` goes to infinity.

    Requires this morphism to be an endomorphism.

    EXAMPLES::

        sage: WordMorphism('a->abcd,b->cd,c->dd,d->').immortal_letters()
        {'a'}
    """
    if not self.is_endomorphism():
        raise TypeError(f'self ({self}) is not an endomorphism')

    forward, backward = {}, {}
    for letter in self._morph:
        forward[letter], backward[letter] = set(), set()

    stack = []
    for preimage, image in self._morph.items():
        if not image:
            stack.append(preimage)
        for occurrence in image:
            forward[preimage].add(occurrence)
            backward[occurrence].add(preimage)

    while stack:
        letter = stack.pop()
        for preimage in backward[letter]:
            forward[preimage].remove(letter)
            if not forward[preimage]:
                stack.append(preimage)
        del forward[letter]
        del backward[letter]

    return set(forward)

def growing_letters(self):
    r"""
    Return the list of growing letters.

    See :meth:`.is_growing` for more information.

    EXAMPLES::

        sage: WordMorphism('0->01,1->10').growing_letters()
        ['0', '1']
        sage: WordMorphism('0->01,1->1').growing_letters()
        ['0']
        sage: WordMorphism('0->01,1->0,2->1',codomain=Words('012')).growing_letters()
        ['0', '1', '2']

    TESTS:

    Make sure that :trac:`31454` is fixed::

        sage: WordMorphism('a->a').growing_letters()
        []
    """
    # Remove letters of type b->a, a->.
    immortal = self.immortal_letters()
    new_morph = {x : [z for z in y if z in immortal]
            for x, y in self._morph.items() if x in immortal}

    # Remove letters of type c->bd, d->ca.
    graph_first = {x : y[0] for x, y in new_morph.items()}
    loops = set()
    for cycle in get_cycles(graph_first.__getitem__, graph_first):
        if any(len(new_morph[letter]) > 1 for letter in cycle):
            continue
        loops.update(cycle)
    new_morph = {x : [z for z in y if z not in loops]
            for x, y in new_morph.items() if x not in loops}

    # Remove letters of type e->abcd.
    new_morph = WordMorphism(new_morph, domain=self.domain(), codomain=self.codomain())
    result = new_morph.immortal_letters()

    return sorted(result, key=self.domain().alphabet().rank)

def is_growing(self, letter=None):
    r"""
    Return ``True`` if ``letter`` is a growing letter.

    A letter `a` is *growing* for the morphism `s` if the length of the
    iterates of `| s^n(a) |` tend to infinity as `n` goes to infinity.

    INPUT:

    - ``letter`` -- ``None`` or a letter in the domain of ``self``

    .. NOTE::

        If letter is ``None``, this returns ``True`` if ``self`` is
        everywhere growing, i.e., all letters are growing letters (see
        [CassNic10]_), and that ``self`` **must** be an endomorphism.

    EXAMPLES::

        sage: WordMorphism('0->01,1->1').is_growing('0')
        True
        sage: WordMorphism('0->01,1->1').is_growing('1')
        False
        sage: WordMorphism('0->01,1->10').is_growing()
        True
        sage: WordMorphism('0->1,1->2,2->01').is_growing()
        True
        sage: WordMorphism('0->01,1->1').is_growing()
        False

    The domain needs to be equal to the codomain::

        sage: WordMorphism('0->01,1->0,2->1',codomain=Words('012')).is_growing()
        True

    Test of erasing morphisms::

        sage: WordMorphism('0->01,1->').is_growing('0')
        False
        sage: m = WordMorphism('a->bc,b->bcc,c->',codomain=Words('abc'))
        sage: m.is_growing('a')
        False
        sage: m.is_growing('b')
        False
        sage: m.is_growing('c')
        False

    TESTS:

    Make sure that :trac:`31454` is fixed::

        sage: WordMorphism('a->a').is_growing('a')
        False

    REFERENCES:

    ..  [CassNic10] Cassaigne J., Nicolas F. Factor complexity.
        Combinatorics, automata and number theory, 163--247, Encyclopedia
        Math. Appl., 135, Cambridge Univ. Press, Cambridge, 2010.
    """
    if not letter:
        return self.domain().alphabet().cardinality() == len(self.growing_letters())
    else:
        return letter in self.growing_letters()

# minimal_conjugate, is_injective, simplify, infinite_repetitions... - all new code.
# https://trac.sagemath.org/ticket/18119

def minimal_conjugate(self):
    r"""
    Return the lexicographically minimal conjugate of this word (see
    :wikipedia:`Lexicographically_minimal_string_rotation`).

    EXAMPLES::

        sage: Word('213').minimal_conjugate()
        word: 132
        sage: Word('11').minimal_conjugate()
        word: 11
        sage: Word('12112').minimal_conjugate()
        word: 11212
        sage: Word('211').minimal_conjugate()
        word: 112
        sage: Word('211211211').minimal_conjugate()
        word: 112112112

    TESTS::

        sage: Word().minimal_conjugate()
        word:
    """
    if not self:
        return self
    p = self.primitive()
    q = self.length() // p.length()
    end = 0
    for factor in (p ** 2).lyndon_factorization():
        end += factor.length()
        if end >= p.length():
            return factor ** q

def is_injective(self):
    """
    Return whether this morphism is injective.

    ALGORITHM:

    Uses a version of :wikipedia:`Sardinas–Patterson_algorithm`.
    Time complexity is on average quadratic with regards to the size of the
    morphism.

    EXAMPLES::

        sage: WordMorphism('a->0,b->10,c->110,d->111').is_injective()
        True
        sage: WordMorphism('a->00,b->01,c->012,d->20001').is_injective()
        False
    """
    def check(u, v):
        if u.is_prefix(v):
            tail = v[u.length():]
            if tail not in tails:
                tails.add(tail)
                todo.append(tail)

    if self.is_erasing():
        return False
    images = self.images()
    tails = set()
    todo = []

    for i in range(len(images)):
        for j in range(i + 1, len(images)):
            if images[i] == images[j]:
                return False
            check(images[i], images[j])
            check(images[j], images[i])
    while todo:
        u = todo.pop()
        for v in images:
            if u == v:
                return False
            check(u, v)
            check(v, u)

    return True

def is_pushy(self, w=None):
    r"""
    Return whether the language `\{m^n(w) | n \ge 0\}` is pushy,
    where `m` is this morphism and `w` is a word inputted as a parameter.

    Requires this morphism to be an endomorphism.

    A language created by iterating a morphism is pushy if its words
    contain an infinite number of factors containing no growing letters. It
    turns out that this is equivalent to having at least one infinite
    repetition containing no growing letters.

    See :meth:`infinite_repetitions` and :meth:`is_growing`.

    INPUT:

    - ``w`` -- finite iterable representing a word used to start the
      language, default is ``self.domain().alphabet()``

    EXAMPLES::

        sage: WordMorphism('a->abca,b->bc,c->').is_pushy()
        False
        sage: WordMorphism('a->abc,b->,c->bcb').is_pushy()
        True
    """
    return bool(self.infinite_repetitions_bounded(w))

def is_unboundedly_repetitive(self, w=None):
    r"""
    Return whether the language `\{m^n(w) | n \ge 0\}` is unboundedly repetitive,
    where `m` is this morphism and `w` is a word inputted as a parameter.

    Requires this morphism to be an endomorphism.

    A language created by iterating a morphism is unboundedly repetitive if
    it has at least one infinite repetition containing at least one growing
    letter.

    See :meth:`infinite_repetitions` and :meth:`is_growing`.

    INPUT:

    - ``w`` -- finite iterable representing a word used to start the
      language, default is ``self.domain().alphabet()``

    EXAMPLES::

        sage: WordMorphism('a->abca,b->bc,c->').is_unboundedly_repetitive()
        True
        sage: WordMorphism('a->abc,b->,c->bcb').is_unboundedly_repetitive()
        False
    """
    return bool(self.infinite_repetitions_growing(w))

def is_repetitive(self, w=None):
    r"""
    Return whether the language `\{m^n(w) | n \ge 0\}` is repetitive,
    where `m` is this morphism and `w` is a word inputted as a parameter.

    Requires this morphism to be an endomorphism.

    A language is repetitive if for each positive integer `k` there exists
    a word `u` such that `u^k` is a factor of some word of the language.

    It turns that for languages created by iterating a morphism this is
    equivalent to having at least one infinite repetition (this property is
    also known as strong repetitiveness).

    See :meth:`infinite_repetitions`.

    INPUT:

    - ``w`` -- finite iterable representing a word used to start the
      language, default is ``self.domain().alphabet()``

    EXAMPLES:

    This method can be used to check whether a purely morphic word is NOT
    k-power free for all positive integers k. For example, the language
    containing just the Thue-Morse word and its prefixes is not repetitive,
    since the Thue-Morse word is cube-free::

        sage: WordMorphism('a->ab,b->ba').is_repetitive('a')
        False

    Similarly, the Hanoi word is square-free::

        sage: WordMorphism('a->aC,A->ac,b->cB,B->cb,c->bA,C->ba').is_repetitive('a')
        False

    However, this method solves a more general problem, as it can be called
    on any morphism `m` and with any word `w`::

        sage: WordMorphism('a->c,b->cda,c->a,d->abc').is_repetitive('bd')
        True
    """
    return self.is_pushy(w) or self.is_unboundedly_repetitive(w)

def infinite_repetitions(self, w=None):
    r"""
    Return the set of primitive infinite repetitions (up to conjugacy)
    from the language `\{m^n(w) | n \ge 0\}`, where `m` is this morphism
    and `w` is a word inputted as a parameter.

    Requires this morphism to be an endomorphism.

    A non-empty word `v` is an infinite repetition (also known as an
    infinite periodic factor) of a language if for each positive integer
    `k` the word `v^k` is a factor of some word from the language.

    If `v` is an infinite repetition, then all its powers are also infinite
    repetitions, therefore this method returns only the primitive ones. It
    turns out that a language created by iterating a morphism has a finite
    number of primitive infinite repetitions.

    Similarly, if `v` is an infinite repetition, then all its conjugates
    are also infinite repetitions, therefore this method returns only the
    lexicographically minimal one from each conjugacy class.

    INPUT:

    - ``w`` -- finite iterable representing a word used to start the
      language, default is ``self.domain().alphabet()``

    EXAMPLES::

        sage: m = WordMorphism('a->aba,b->aba,c->cd,d->e,e->d')
        sage: inf_reps = m.infinite_repetitions('ac')
        sage: sorted(inf_reps)
        [word: aab, word: de]

    Incomplete check that these words are indeed infinite repetitions::

        sage: SL = m._language_naive(10, Word('ac'))
        sage: all(x in SL for x in inf_reps)
        True
        sage: all(x^2 in SL for x in inf_reps)
        True
        sage: all(x^3 in SL for x in inf_reps)
        True

    Larger example::

        sage: m = WordMorphism('a->1b5,b->fcg,c->dae,d->432,e->678,f->f,g->g,1->2,2->3,3->4,4->1,5->6,6->7,7->8,8->5')
        sage: sorted(m.infinite_repetitions('a'))
        [word: 1432f2143f3214f4321f, word: 5678g8567g7856g6785g]
    """
    return self.infinite_repetitions_bounded(w) | self.infinite_repetitions_growing(w)

def infinite_repetitions_bounded(self, w=None):
    r"""
    Return the set of primitive infinite repetitions (up to conjugacy),
    which contain no growing letters,
    from the language `\{m^n(w) | n \ge 0\}`, where `m` is this morphism
    and `w` is a word inputted as a parameter.

    Requires this morphism to be an endomorphism.

    See :meth:`infinite_repetitions` and :meth:`is_growing`.

    INPUT:

    - ``w`` -- finite iterable representing a word used to start the
      language, default is ``self.domain().alphabet()``

    ALGORITHM:

    The algorithm used is described in detail in [KS2015]_.

    EXAMPLES::

        sage: m = WordMorphism('a->aba,b->aba,c->cd,d->e,e->d')
        sage: sorted(m.infinite_repetitions_bounded())
        [word: de]

        sage: m = WordMorphism('c->d,d->c,e->fc,f->ed')
        sage: sorted(m.infinite_repetitions_bounded())
        [word: c, word: d]

    TESTS::

        sage: m = WordMorphism('a->Cab,b->1c1,c->E2bd5,d->BbaA,5->6,6->7,7->8,8->9,9->5,1->2,2->1,A->B,B->C,C->D,D->E,E->')
        sage: sorted(m.infinite_repetitions_bounded())
        [word: 1, word: 1519181716, word: 2, word: 2529282726]

        sage: WordMorphism('a->b,b->b', codomain=FiniteWords('ab')).infinite_repetitions()
        set()
    """
    def impl():
        U = {}
        for x in unbounded:
            xg = g.image(x)
            for i, y in enumerate(reversed(xg)):
                if y in unbounded:
                    break
            U[x] = y, xg[xg.length() - i:]
        for cycle in get_cycles(lambda x: U[x][0], domain=unbounded):
            if all(not U[x][1] for x in cycle):
                continue
            gq = gb ** len(cycle)
            for cycle in g.domain()(cycle).conjugates_iterator():
                u = g.domain()()
                for x in cycle:
                    u = U[x][1] + gb(u)
                inf_rep = g.domain()()
                history = set()
                while u not in history:
                    history.add(u)
                    inf_rep += u
                    u = gq(u)
                yield k(inf_rep.primitive()).primitive()

    if w is None:
        w = self._morph
    f = self.restrict_domain(self.reach(w))
    f._codomain = f._domain
    g, _, k, _ = f.simplify_injective()
    g._codomain = g._domain
    unbounded = set(g.growing_letters())
    gb = g.restrict_domain(set(g._morph) - unbounded)

    result = set()
    for x in impl():  # UR.
        result.add(x.minimal_conjugate())
    g, k = g.reversal(), k.reversal()
    for x in impl():  # UL.
        result.add(self.domain()(reversed(x)).minimal_conjugate())

    return result

def infinite_repetitions_growing(self, w=None):
    r"""
    Return the set of primitive infinite repetitions (up to conjugacy),
    which contain at least one growing letter,
    from the language `\{m^n(w) | n \ge 0\}`, where `m` is this morphism
    and `w` is a word inputted as a parameter.

    Requires this morphism to be an endomorphism.

    See :meth:`infinite_repetitions` and :meth:`is_growing`.

    INPUT:

    - ``w`` -- finite iterable representing a word used to start the
      language, default is ``self.domain().alphabet()``

    ALGORITHM:

    The algorithm used is described in detail in [KS2015]_.

    EXAMPLES::

        sage: m = WordMorphism('a->aba,b->aba,c->cd,d->e,e->d')
        sage: sorted(m.infinite_repetitions_growing())
        [word: aab]

        sage: m = WordMorphism('a->bcb,b->ada,c->d,d->c')
        sage: sorted(m.infinite_repetitions_growing())
        [word: ad, word: bc]

        sage: m = WordMorphism('b->c,c->bcb')
        sage: sorted(m.infinite_repetitions_growing())
        [word: bc]

        sage: m = WordMorphism('a->abc,b->dab,c->abc,d->dab')
        sage: sorted(m.infinite_repetitions_growing())
        [word: ababcd]
    """
    if w is None:
        w = self._morph
    f = self.restrict_domain(self.reach(w))
    f._codomain = f._domain
    g, _, k, _ = f.simplify_injective()
    g._codomain = g._domain
    unbounded = set(g.growing_letters())

    result = set()
    for periodic_orbit in g.periodic_points():
        gq = g ** len(periodic_orbit)
        for periodic_point in periodic_orbit:
            # Check if this periodic point is a periodic infinite word.
            periodic_point = periodic_point[:1]
            occurred = set(periodic_point)
            one_unbounded_twice = False
            for _ in g.domain().alphabet():
                previous_length = periodic_point.length()
                periodic_point = gq(periodic_point)
                for i, letter in enumerate(periodic_point[previous_length:]):
                    if letter in unbounded:
                        if letter in occurred:
                            one_unbounded_twice = True
                            break
                        occurred.add(letter)
                if one_unbounded_twice:
                    break
            if not one_unbounded_twice or letter != periodic_point[0]:
                break
            v = periodic_point[:previous_length + i]
            vq = gq(v)
            m = 0
            while vq[m * v.length() : (m + 1) * v.length()] == v:
                m += 1
            if m * v.length() != vq.length():
                break
            result.add(k(v).primitive().minimal_conjugate())

    return result

def reach(self, w):
    r"""
    Return the set of letters which occur in words of
    `\{m^n(w) | n \ge 0\}`, where `m` is this morphism and `w` is a word
    (finite iterable is enough) inputted as a parameter.

    Requires this morphism to be an endomorphism.

    EXAMPLES::

        sage: sorted(WordMorphism('a->ac,b->ce,c->bd,d->d,e->').reach('c'))
        ['b', 'c', 'd', 'e']
    """
    if not self.is_endomorphism():
        raise TypeError(f'self ({self}) is not an endomorphism')

    visited = set(w)
    todo = list(visited)
    while todo:
        a = todo.pop()
        for b in self.image(a):
            if b not in visited:
                visited.add(b)
                todo.append(b)
    return visited

def simplify(self, Z=None):
    r"""
    If this morphism is simplifiable, return morphisms `h` and `k` such that
    this morphism is simplifiable with respect to `h` and `k`, otherwise
    raise  ``ValueError``.

    This method is quite fast if this morphism is non-injective, but very
    slow if it is injective.

    Let `f: X^* \rightarrow Y^*` be a morphism. Then `f` is simplifiable
    with respect to morphisms `h: X^* \rightarrow Z^*` and
    `k: Z^* \rightarrow Y^*`, if `f = k \circ h` and `|Z| < |X|`. If also
    `Y \subseteq X`, then the morphism `g: Z^* \rightarrow Z^* = h \circ k`
    is a simplification of `f` (with respect to `h` and `k`).

    Loosely speaking a morphism is simplifiable if it contains "more letters
    than is needed". Non-injectivity implies simplifiability. Simplification
    preserves some properties of the original morphism (e.g. repetitiveness).

    For more information see Section 3 in [KO2000]_.

    INPUT:

    - ``Z`` -- iterable, whose elements are used as an alphabet for the
      simplification, default is ``self.domain().alphabet()``

    EXAMPLES:

    Example of a simplifiable (non-injective) morphism::

        sage: f = WordMorphism('a->aca,b->badc,c->acab,d->adc')
        sage: h, k = f.simplify('xyz'); h, k
        (WordMorphism: a->x, b->zy, c->xz, d->y, WordMorphism: x->aca, y->adc, z->b)
        sage: k * h == f
        True
        sage: g = h * k; g
        WordMorphism: x->xxzx, y->xyxz, z->zy

    Example of a simplifiable (injective) morphism::

        sage: f = WordMorphism('a->abcc,b->abcd,c->abdc,d->abdd')
        sage: h, k = f.simplify('xyz'); h, k
        (WordMorphism: a->xyy, b->xyz, c->xzy, d->xzz, WordMorphism: x->ab, y->c, z->d)
        sage: k * h == f
        True
        sage: g = h * k; g
        WordMorphism: x->xyyxyz, y->xzy, z->xzz

    Example of a non-simplifiable morphism::

        sage: WordMorphism('a->aa').simplify()
        Traceback (most recent call last):
        ...
        ValueError: self (a->aa) is not simplifiable

    Example of an erasing morphism::

        sage: f = WordMorphism('a->abc,b->cc,c->')
        sage: h, k = f.simplify(); h, k
        (WordMorphism: a->a, b->b, c->, WordMorphism: a->abc, b->cc)
        sage: k * h == f
        True
        sage: g = h * k; g
        WordMorphism: a->ab, b->

    Example of a morphism, that is not an endomorphism::

        sage: f = WordMorphism('a->xx,b->xy,c->yx,d->yy')
        sage: h, k = f.simplify(NN); h, k
        (WordMorphism: a->00, b->01, c->10, d->11, WordMorphism: 0->x, 1->y)
        sage: k * h == f
        True
        sage: len(k.domain().alphabet()) < len(f.domain().alphabet())
        True
    """
    def try_create_h(f, k):
        h = {}
        for letter1, image1 in f.items():
            image3 = []
            while image1:
                for letter2, image2 in k.items():
                    if image2.is_prefix(image1):
                        image1 = image1[image2.length():]
                        image3.append(letter2)
                        break
                else:  # nobreak
                    return None
            h[letter1] = image3
        return h

    X = self.domain().alphabet()
    Y = self.codomain().alphabet()
    f = self._morph

    if self.is_erasing():  # Trivial case #1.
        k = {letter: image for letter, image in f.items() if image}
        h = {letter: [letter] if image else [] for letter, image in f.items()}
    elif len(Y) < len(X):  # Trivial case #2.
        k = {x: [y] for x, y in zip(X, Y)}
        k_inverse = {y: x for y, x in zip(Y, X)}
        h = {x: [k_inverse[y] for y in image] for x, image in f.items()}
    elif not self.is_injective():  # Non-trivial but a fast case.
        k = dict(f)
        to_do = set(k)
        while to_do:
            to_remove = []
            # min() and remove() instead of pop() to have deterministic output.
            letter1 = min(to_do)
            to_do.remove(letter1)
            image1 = k[letter1]
            for letter2, image2 in k.items():
                if letter1 == letter2:
                    continue
                if image1 == image2:
                    to_remove.append(letter2)
                    to_do.discard(letter2)
                elif image1.is_prefix(image2):
                    k[letter2] = image2[image1.length():]
                    to_do.add(letter2)
                elif image2.is_prefix(image1):
                    k[letter1] = image1[image2.length():]
                    to_do.add(letter1)
                    break
            for letter in to_remove:
                del k[letter]
        h = try_create_h(f, k)
    else:  # Non-trivial and a slow case.
        factors = set()
        for image in f.values():
            factors.update(x.primitive() for x in image.factor_iterator())
        factors.remove(self.codomain()())
        factors = sorted(factors)  # For deterministic output.
        from itertools import combinations
        for comb in combinations(factors, len(X) - 1):
            if any(x.is_proper_prefix(y) for x in comb for y in comb):
                continue
            k = {x: image for x, image in zip(X, comb)}
            h = try_create_h(f, k)
            if h:
                break
        else:  # nobreak
            raise ValueError(f'self ({self}) is not simplifiable')

    k = WordMorphism(k, codomain=self.codomain())
    h = WordMorphism(h, domain=self.domain(), codomain=k.domain())

    if Z is not None:  # Custom alphabet.
        old_Z_star = k.domain()
        old_Z = old_Z_star.alphabet()
        Z = [z for z, _ in zip(Z, old_Z)]
        if len(Z) < len(old_Z):
            raise ValueError(f'Z should have length at least {len(old_Z)}, is {len(Z)}')
        Z_star = FiniteWords(Z)
        h_new = {old: [new] for old, new in zip(old_Z, Z)}
        k_new = {new: [old] for new, old in zip(Z, old_Z)}
        h_new = WordMorphism(h_new, domain=old_Z_star, codomain=Z_star)
        k_new = WordMorphism(k_new, domain=Z_star, codomain=old_Z_star)
        h = h_new * h
        k = k * k_new

    return h, k

def simplify_injective(self):
    r"""
    Return a quadruplet `(g, h, k, i)`, where `g` is an injective
    simplification of this morphism with respect to `h`, `k` and `i`.

    Requires this morphism to be an endomorphism.

    This methods basically calls :meth:`simplify` until the returned
    simplification is injective. If this morphism is already injective, a
    quadruplet `(g, h, k, i)` is still returned, where `g` is this morphism,
    `h` and `k` are the identity morphisms and `i` is 0.

    Let `f: X^* \rightarrow Y^*` be a morphism and `Y \subseteq X`. Then
    `g: Z^* \rightarrow Z^*` is an injective simplification of `f` with
    respect to morphisms `h: X^* \rightarrow Z^*` and
    `k: Z^* \rightarrow Y^*` and a positive integer `i`, if `g` is
    injective, `|Z| < |X|`, `g^i = h \circ k` and `f^i = k \circ h`.

    For more information see Section 4 in [KO2000]_.

    EXAMPLES::

        sage: f = WordMorphism('a->abc,b->a,c->bc')
        sage: g, h, k, i = f.simplify_injective(); g, h, k, i
        (WordMorphism: a->aa, WordMorphism: a->aa, b->a, c->a, WordMorphism: a->abc, 2)
        sage: g.is_injective()
        True
        sage: g ** i == h * k
        True
        sage: f ** i == k * h
        True
    """
    if not self.is_endomorphism():
        raise TypeError(f'self ({self}) is not an endomorphism')

    g = self
    h = self.domain().identity_morphism()
    k = self.codomain().identity_morphism()
    i = 0
    while not g.is_injective():
        h_new, k_new = g.simplify()
        g, h, k, i = h_new * k_new, h_new * h, k * k_new, i + 1
    return g, h, k, i

# Partial rewrite of _language_naive.
# https://trac.sagemath.org/ticket/31684

def language_naive(self, n, u):
    r"""
    Return all words of length less than ``n`` by naive substitution.

    The language of the substitution is the DOL language which consist
    of factors of `s^n(u)`.

    This method assumes this substitution is non-erasing.

    INPUT:

    - ``n`` -- non-negative integer - length of the words in the language

    - ``u`` -- a word used as a seed

    OUTPUT: a Python set

    TESTS::

        sage: s = WordMorphism({0: [0,1], 1:[0]})
        sage: W = s.domain()
        sage: sorted(s._language_naive(3, W([0])))
        [word: 0, word: 00, word: 01, word: 1, word: 10]
        sage: sorted(s._language_naive(3, W([1])))
        [word: 0, word: 00, word: 01, word: 1, word: 10]

        sage: s._language_naive(3, W())
        set()
        sage: W([1, 1]) in s._language_naive(3, W([1, 1]))
        True
    """
    L = set()
    todo = []
    for i in range(len(u)):
        for j in range(i+1, min(len(u)+1, i+n)):
            f = u[i:j]
            if f not in L:
                todo.append(f)
                L.add(f)
    while todo:
        u = todo.pop()
        v = self(u)
        if u.length() == 1:
            for i in range(len(v)):
                for j in range(i+1, min(len(v)+1, i+n)):
                    f = v[i:j]
                    if f not in L:
                        todo.append(f)
                        L.add(f)
        else:
            l = self._morph[u[0]].length()
            r = self._morph[u[-1]].length()
            m = v.length() - l - r
            x = n - 1 - m
            for i in range(l - min(x - 1, l), l):
                for j in range(l + m + 1, l + m + 1 + min(x - l + i, r)):
                    f = v[i:j]
                    if f not in L:
                        todo.append(f)
                        L.add(f)
    return L

# Some tiny bugfixes.
# https://trac.sagemath.org/ticket/31454

from sage.misc.callable_dict import CallableDict
from sage.combinat.words.morphism import PeriodicPointIterator
def _periodic_points_fix(self):
    r"""
    ...
    """
    assert self.is_endomorphism(), "f should be an endomorphism"

    if self.is_erasing():
        raise NotImplementedError("f should be non erasing")

    A = self.domain().alphabet()
    d = dict((letter,self(letter)[0]) for letter in A)
    G = set(self.growing_letters())

    res = []
    parent = self.codomain().shift()
    for cycle in get_cycles(CallableDict(d),A):
        if cycle[0] in G:
            P = PeriodicPointIterator(self, cycle)
            res.append([parent(P._cache[i]) for i in range(len(cycle))])

    return res

def _growing_letters_fix(self):
    r"""
    ...
    """
    if self.is_primitive() and len(self._morph) > 1:
        return self.domain().alphabet().list()
    last_coef = 0
    coefs = self.incidence_matrix().charpoly().coefficients(sparse=False)
    while coefs[last_coef] == 0:
        last_coef += 1
    V = self.abelian_rotation_subspace() + (self.incidence_matrix()**last_coef).right_kernel().change_ring(QQ)
    basis = V.ambient_vector_space().basis()
    A = self.domain().alphabet()

    return list(A.unrank(i) for i in range(A.cardinality()) if basis[i] not in V)

# Monkey patching.
from sage.combinat.words.morphism import get_cycles, WordMorphism
from sage.combinat.words.finite_word import FiniteWord_class
# Bug fixes.
WordMorphism.periodic_points = _periodic_points_fix
WordMorphism.growing_letters = _growing_letters_fix
# Save old methods for testing.
WordMorphism.growing_letters_old = WordMorphism.growing_letters
WordMorphism._language_naive_old = WordMorphism._language_naive
# New methods.
FiniteWord_class.minimal_conjugate = minimal_conjugate
WordMorphism.immortal_letters = immortal_letters
WordMorphism.growing_letters = growing_letters
WordMorphism.is_growing = is_growing
WordMorphism.is_injective = is_injective
WordMorphism.is_pushy = is_pushy
WordMorphism.is_unboundedly_repetitive = is_unboundedly_repetitive
WordMorphism.is_repetitive = is_repetitive
WordMorphism.infinite_repetitions = infinite_repetitions
WordMorphism.infinite_repetitions_bounded = infinite_repetitions_bounded
WordMorphism.infinite_repetitions_growing = infinite_repetitions_growing
WordMorphism.reach = reach
WordMorphism.simplify = simplify
WordMorphism.simplify_injective = simplify_injective
WordMorphism._language_naive = language_naive
