import random
from itertools import product
from functools import wraps
from sage.combinat.words.morphism import WordMorphism
from sage.combinat.words.words import FiniteWords
from sage.combinat.words.word import Word

# Bonus methods:
def necessary_condition_for_elementary(self):
    alph = [set(image) for image in self._morph.values()]
    for x in product(*alph):
        if len(set(x)) == len(x):
            return True
    return False

def is_injective_with_cert(self, certificate=False):
    r"""
    Return whether this morphism is injective.

    EXAMPLES::
        sage: WordMorphism('a->0,b->10,c->110,d->111').is_injective()
        True
        sage: WordMorphism('a->00,b->01,c->012,d->20001').is_injective()
        False

    One pair of preimages mapped to the same word can be obtained by
    setting the ``certificate`` parameter to ``True``::

        sage: m = WordMorphism('a->00,b->01,c->012,d->20001')
        sage: answer, u, v = m.is_injective(certificate=True); answer, u, v
        (False, word: bd, word: cab)
        sage: m(u) == m(v)
        True

    If the morphism is injective, ``None`` is returned instead::

        sage: WordMorphism('a->0,b->10,c->110,d->111').is_injective(True)
        True, None, None
    """
    domain = self.domain()
    for letter, image in self._morph.items():
        if not image:
            return False if not certificate else (False, domain([letter]), domain([letter, letter]))

    todo = []
    tails = {}

    for letter1, image1 in self._morph.items():
        for letter2, image2 in self._morph.items():
            if letter1 == letter2:
                continue
            elif image1.is_prefix(image2):
                tail = image2[image1.length():]
                if tail not in tails:
                    todo.append(tail)
                    tails[tail] = ([letter1], [letter2])

    while todo:
        tail = todo.pop()
        for letter, image in self._morph.items():
            if tail == image:
                preimage1, preimage2 = tails[tail]
                return False if not certificate else (False, domain(preimage2), domain(preimage1 + [letter]))
            elif tail.is_prefix(image):
                new_tail = image[tail.length():]
                if new_tail not in tails:
                    todo.append(new_tail)
                    preimage1, preimage2 = tails[tail]
                    tails[new_tail] = (preimage2, preimage1 + [letter])
            elif image.is_prefix(tail):
                new_tail = tail[image.length():]
                if new_tail not in tails:
                    todo.append(new_tail)
                    preimage1, preimage2 = tails[tail]
                    tails[new_tail] = (preimage1 + [letter], preimage2)

    return True if not certificate else (True, None, None)

def geometric_sample_with_limit(start, end, p):
    i = start
    while i < end and random.random() > p:
        i += 1
    return i

# a1 - min size of alphabet
# a2 - max size of alphabet
# i1 - min length of image
# i2 - max length of image
# ip - probability that image has minimal length
def create_random_morphism(a1=3, a2=8, i1=0, i2=10, ip=0.2):
    m = dict()
    X_size = random.randint(a1, a2)
    X = list(range(X_size)) # string.ascii_lowercase[:X_size]
    for letter in X:
        image_size = geometric_sample_with_limit(i1, i2, ip)
        m[letter] = random.choices(X, k=image_size)
    return WordMorphism(m)

# Decorator to save on some boilerplate:
def _test(logger=[0], a1=3, a2=8, i1=0, i2=10, ip=0.2, cnt=10000, seed=18):
    def decorator(f):
        @wraps(f)
        def wrapper(a1=a1, a2=a2, i1=i1, i2=i2, ip=ip, cnt=cnt, seed=seed, debug=False):
            random.seed(seed)
            for i in range(cnt):
                print(f'\r{i}' + (' ' * 50), end='')
                m = create_random_morphism(a1, a2, i1, i2, ip)
                if debug: print(f'\n{m}')
                f(m, debug, logger)
            print(f'\r{cnt}')
            return [x / cnt for x in logger]
        return wrapper
    return decorator

# Tests as described in the text of thesis:
@_test(logger=[0, 0, 0], a1=3, a2=50, i1=0, i2=5, ip=0.5)
def test_growing_letters(m, debug, logger):
    m._codomain = m._domain

    res0 = m.immortal_letters()
    res1 = m.growing_letters()
    if debug: print(res0)
    if debug: print(res1)

    logger[0] += (len(m._morph) - len(res0)) / len(m._morph)
    logger[1] += (len(res0) - len(res1)) / len(m._morph)
    logger[2] += len(res1) / len(m._morph)

    res2 = m.growing_letters_old()

    if res1 != res2:
        print(m)
        print(res1)
        print(res2)
        assert(False)

@_test(logger=[0, 0])
def test_is_injective(m, debug, logger):
    res1 = m.is_injective()
    if debug: print(is_injective1)

    res2, u, v = is_injective_with_cert(m, True)
    assert(res1 == res2)
    if not res2:
        logger[0] += 1
        if len(u) + len(v) > 4:
            logger[1] += 1
        if m(u) != m(v):
            print(m)
            print(u)
            print(v)
            assert(False)

@_test(logger=[0, 0], a1=3, a2=8, i1=1, i2=3, ip=0.5)
def test_simplify(f, debug, logger):
    try:
        h, k = f.simplify(NN)
    except ValueError:
        assert(f.is_injective())
        assert(necessary_condition_for_elementary(f))
        logger[0] += 1
        logger[1] += 1
        return
    g = h * k
    assert(f == k * h)
    assert(len(g._domain.alphabet()) < len(f._domain.alphabet()))
    if f.is_injective():
        logger[1] += 1

@_test(logger=[0, 0])
def test_simplify_injective(f, debug, logger):
    f._codomain = f._domain

    g, h, k, i = f.simplify_injective()
    if i == 0:
        if debug: print('failed to simplify')
        assert(f.is_injective())
        return
    logger[0] += 1
    if i > 1:
        logger[1] += 1
    if debug: print(f'g: {g}\nh: {h}\nk: {k}\ni: {i}')
    assert(len(g._domain.alphabet()) < len(f._domain.alphabet()))
    assert(g.is_injective())
    gi = g ** i
    gi2 = h * k
    if debug: print(f'gi: {gi}\ngi?: {gi2}')
    assert(gi == gi2)
    fi = f ** i
    fi2 = k * h
    if debug: print(f'fi: {fi}\nfi?: {fi2}')
    assert(fi == fi2)

@_test(a1=3, a2=15, i1=0, i2=2, ip=1/3)
def test_reach(m, debug, logger):
    m._codomain = m._domain

    alph = m.domain().alphabet()
    word_size = random.randint(1, 3)
    w = random.choices(alph, k=word_size)
    if debug: print(w)
    res1 = m.reach(w)
    if debug: print(res1)
    logger[0] += len(res1) / len(alph)

    v = w
    res2 = set(v)
    for _ in alph:
        v = m(v)
        res2.update(v)

    if res1 != res2:
        print(m)
        print(w)
        print(res1)
        print(res2)
        assert(False)

@_test(logger=[0, 0, 0, 0, 0, 0])
def test_infinite_repetitions(f, debug, logger):
    f._codomain = f._domain

    g, _, _, _ = f.simplify_injective()
    g._codomain = g._domain

    limit = 20+1
    SL = g._language_naive(limit, g.domain()(list(g._morph)))
    growing = set(g.growing_letters())
    inf_reps = g.infinite_repetitions()
    for inf_rep in inf_reps:
        if all(letter not in growing for letter in inf_rep):
            logger[0] += 1
            logger[4] += 1
            if len(inf_rep) > 1:
                logger[1] += 1
        else:
            logger[2] += 1
            if len(inf_rep) > 1:
                logger[3] += 1
            logger[4] += 1
        for i in range(2, limit):
            q = len(inf_rep) * i
            if q >= limit:
                break
            logger[5] += 1
            if inf_rep ** i not in SL:
                print(f)
                print(inf_rep, i)
                assert(False)

@_test(a1=3, a2=8, i1=1, i2=5, ip=1/3)
def test_language_naive(m, debug, logger):
    w = m.domain().random_element()
    if debug: print(w)
    res1 = m._language_naive(10, w)
    logger[0] += len(res1)

    m._morph[255] = w
    res2 = m._language_naive_old(10, Word([255]))

    if res1 != res2:
        print(m)
        print(w)
        print(res1 - res2)
        print(res2 - res1)
        assert(False)

@_test(a1=10, a2=10)
def test_minimal_conjugate(m, debug, logger):
    w = m.domain().random_element()
    if random.randint(0, 1):
        w = w ** random.randint(2, 5)

    res1 = w.minimal_conjugate()
    if debug: print(res1)
    if w.is_primitive():
        logger[0] += 1

    res2 = w
    for v in w.conjugates_iterator():
        if v < res2:
            res2 = v

    if res1 != res2:
        print(f'w: {w}')
        print(f'res1: {res1}')
        print(f'res2: {res2}')
        assert(False)
